#!/usr/bin/env python3

import cairo
from gi.repository import Gtk, Gdk,GdkPixbuf

class ComipoliCanvas (Gtk.DrawingArea):
    def __init__(self):
        Gtk.DrawingArea.__init__(self)
        self.first_page = None
        self.second_page = None
        self.l_to_r = False
        self.spread = False
        self.is_pdf = False
        #self.connect('draw', self.on_draw)
        self.set_draw_func(self.draw_func)

    def put(self, page, cr, w, h):
        if self.is_pdf:
            cr.set_source_rgb(1, 1, 1)
            cr.rectangle(0, 0, w, h)
            cr.fill()
            page.render(cr)
        else:
            Gdk.cairo_set_source_pixbuf(cr, page, 0, 0)
            cr.paint()

    def draw_func(self, widget, cr, width, height):
        aw = widget.get_allocated_width()
        ah = widget.get_allocated_height()
        cr.set_source_rgb(0.0, 0.0, 0.0)
        cr.rectangle(0, 0, aw, ah)
        cr.fill()
        if self.first_page != None:
            if self.is_pdf:
                w, h = self.first_page.get_size()
            else:
                w = self.first_page.get_width()
                h = self.first_page.get_height()
            # Horizontal?
            if w - h > 0 or not self.spread:
                # Single Page
                if aw * h > ah * w:
                    n = ah/h
                    matrix = cairo.Matrix(n, 0, 0, n, aw/2-w*n/2, 0)
                    cr.transform(matrix)
                    self.put(self.first_page, cr, w, h)
                else:
                    n = aw/w
                    matrix = cairo.Matrix(n, 0, 0, n, n, ah/2-h*n/2)
                    cr.transform(matrix)
                    self.put(self.first_page, cr, w, h)
            else:
                if self.second_page != None:
                    n = ah/h
                    if self.l_to_r:
                        cr.save()
                        matrix = cairo.Matrix(n, 0, 0, n, aw/2-w*n, n)
                        cr.transform(matrix)
                        self.put(self.first_page, cr, w, h)
                        # Reset
                        cr.restore()
                        # Left Page
                        if self.is_pdf:
                            w, h = self.second_page.get_size()
                        else:
                            w = self.second_page.get_width()
                            h = self.second_page.get_height()
                        n = ah/h
                        matrix = cairo.Matrix(n, 0, 0, n, aw/2, n)
                        cr.transform(matrix)
                        self.put(self.second_page, cr, w, h)
                    else:
                        # Save Matrix
                        cr.save()
                        matrix = cairo.Matrix(n, 0, 0, n, aw/2, n)
                        cr.transform(matrix)
                        self.put(self.first_page, cr, w, h)
                        # Reset
                        cr.restore()
                        # Left Page
                        if self.is_pdf:
                            w, h = self.second_page.get_size()
                        else:
                            w = self.second_page.get_width()
                            h = self.second_page.get_height()
                        n = ah/h
                        matrix = cairo.Matrix(n, 0, 0, n, aw/2-w*n, n)
                        cr.transform(matrix)
                        self.put(self.second_page, cr, w, h)
        return False

