#!/usr/bin/env python3

from gi.repository import Gtk, Gio, GLib, Gdk, GdkPixbuf, GObject

import comipoli_numsort, html, cairo
from comipoli_thumbnail import ComipoliThumbnail
from comipoli_headerbar import ComipoliHeaderBar
from comipoli_archive import ComipoliArchive
from comipoli_canvas import ComipoliCanvas
from comipoli_upperbar import ComipoliUpperBar
#from comipoli_message import ComipoliMessage
from comipoli_menubutton import ComipoliMenuButton
from comipoli_nextdialog import ComipolNextDialog

class ComipoliWindow(Gtk.ApplicationWindow):
    def __init__(self, app):
        Gtk.ApplicationWindow.__init__(self, application=app)
        # var
        self.app = app
        self.thumbnail_height = 120
        self.num = 0
        self.archive = ComipoliArchive()
        self.is_fullscreen = False
        self.long_keypress = 0
        self.long_click = 0
        self.cursor_id = 0
        self.thumbnail_array = []
        # Canvas
        self.canvas = ComipoliCanvas()
        # HeaderBar
        self.headerbar = ComipoliHeaderBar()
        self.headerbar.lr_button.connect('toggled', self.on_lr_button_toggled)
        self.headerbar.grid_button.connect('clicked', self.on_grid_button_clicked)
        self.set_titlebar(self.headerbar)
        # UpperBar
        self.upperbar = ComipoliUpperBar()
        self.upperbar.grid_button.connect('clicked', self.on_grid_button_clicked)
        self.upperbar.restore_button.connect('clicked', self.on_restore_button_clicked)
        GObject.Binding.bind_property(
            self.headerbar.lr_button, 'active',
            self.upperbar.lr_button, 'active',
            GObject.BindingFlags.BIDIRECTIONAL)
        GObject.Binding.bind_property(
            self.headerbar.grid_button, 'sensitive',
            self.upperbar.grid_button, 'sensitive',
            GObject.BindingFlags.BIDIRECTIONAL)
        GObject.Binding.bind_property(
            self.headerbar.title, 'label',
            self.upperbar.title, 'label',
            GObject.BindingFlags.BIDIRECTIONAL)
        GObject.Binding.bind_property(
            self.headerbar.subtitle, 'label',
            self.upperbar.subtitle, 'label',
            GObject.BindingFlags.BIDIRECTIONAL)
        # menu
        self.menu = ComipoliMenuButton()
        self.headerbar.pack_end(self.menu)
        self.menuf = ComipoliMenuButton()
        self.upperbar.pack_end(self.menuf)
        # pack
        overlay = Gtk.Overlay()
        overlay.set_child(self.canvas)
        overlay.add_overlay(self.upperbar)
        self.set_child(overlay)
        # DnD
        drop_target = Gtk.DropTarget.new(Gdk.FileList, Gdk.DragAction.COPY)
        drop_target.connect('drop', self.on_file_drop)
        drop_target.connect('accept', self.on_drop_accept)
        self.add_controller(drop_target)
        # mouse move
        controller_motion = Gtk.EventControllerMotion()
        controller_motion.connect('motion', self.on_mouse_motion)
        self.add_controller(controller_motion)
        # Double Click
        controller_click = Gtk.GestureClick()
        controller_click.connect('released', self.on_cllick_released)
        self.add_controller(controller_click)
        # settings
        self.set_default_size(self.app.win_width, self.app.win_height)
        self.headerbar.lr_button.props.active = self.app.left_to_right
        # show
        self.connect('notify::is-active', self.on_window_active)
        self.connect('notify::default-width', self.on_window_width_change)
        self.connect('notify::default-height', self.on_window_height_change)
        # keyboard
        action_esc = Gio.SimpleAction.new('action_esc', None)
        action_ctrl_w = Gio.SimpleAction.new('action_ctrl_w', None)
        action_space = Gio.SimpleAction.new('action_space', None)
        action_backspace = Gio.SimpleAction.new('action_backspace', None)
        action_ctrl_space = Gio.SimpleAction.new('action_ctrl_space', None)
        action_ctrl_backspace = Gio.SimpleAction.new('action_ctrl_backspace', None)
        action_left = Gio.SimpleAction.new('action_left', None)
        action_right = Gio.SimpleAction.new('action_right', None)
        action_ctrl_left = Gio.SimpleAction.new('action_ctrl_left', None)
        action_ctrl_right = Gio.SimpleAction.new('action_ctrl_right', None)
        action_home = Gio.SimpleAction.new('action_home', None)
        action_end = Gio.SimpleAction.new('action_end', None)
        action_f8 = Gio.SimpleAction.new('action_f8', None)
        action_f9 = Gio.SimpleAction.new('action_f9', None)
        action_f10 = Gio.SimpleAction.new('action_f10', None)
        action_f11 = Gio.SimpleAction.new('action_f11', None)
        #
        action_esc.connect('activate', self.on_action_esc)
        action_ctrl_w.connect('activate', self.on_action_ctrl_w)
        action_space.connect('activate', self.on_action_next, 2)
        action_backspace.connect('activate', self.on_action_next, -2)
        action_ctrl_space.connect('activate', self.on_action_next, 1)
        action_ctrl_backspace.connect('activate', self.on_action_next, -1)
        action_left.connect('activate', self.on_action_arrow, 2)
        action_right.connect('activate', self.on_action_arrow, -2)
        action_ctrl_left.connect('activate', self.on_action_arrow, 1)
        action_ctrl_right.connect('activate', self.on_action_arrow, -1)
        action_home.connect('activate', self.on_action_next, 0)
        action_end.connect('activate', self.on_action_next, 50)
        action_f8.connect('activate', self.on_action_function_key, 8)
        action_f9.connect('activate', self.on_action_function_key, 9)
        action_f10.connect('activate', self.on_action_function_key, 10)
        action_f11.connect('activate', self.on_action_function_key, 11)
        #
        self.add_action(action_esc)
        self.add_action(action_ctrl_w)
        self.add_action(action_space)
        self.add_action(action_backspace)
        self.add_action(action_ctrl_space)
        self.add_action(action_ctrl_backspace)
        self.add_action(action_left)
        self.add_action(action_right)
        self.add_action(action_ctrl_left)
        self.add_action(action_ctrl_right)
        self.add_action(action_home)
        self.add_action(action_end)
        self.add_action(action_f8)
        self.add_action(action_f9)
        self.add_action(action_f10)
        self.add_action(action_f11)

    def on_action_esc(self, action, param):
        if self.is_fullscreen:
            self.toggle_fullscreen()
        elif self.props.application.esc_exit:
            self.close()

    def on_action_ctrl_w(self, action, param):
        self.close()

    def on_action_next(self, action, param, num):
        self.change_pixbuf(num)

    def on_action_arrow(self, action, param, num):
        if self.headerbar.lr_button.props.active:
            self.change_pixbuf(-num)
        else:
            self.change_pixbuf(num)

    def on_action_function_key(self, action, param, num):
        match num:
            case 8:
                self.headerbar.lr_button.props.active = not self.headerbar.lr_button.props.active
            case 9:
                if self.headerbar.grid_button.props.sensitive:
                    self.on_grid_button_clicked(None)
            case 10:
                if self.is_fullscreen:
                    self.upperbar.set_visible(True)()
                    GLib.idle_add(self.show_fullscreen_menu)
                else:
                    self.menu.popup()
            case 11:
                self.toggle_fullscreen()
            case _:
                pass

    def set_uri(self, uri):
        if uri == None:
            return
        if self.archive.new_archive(uri, self.app):
            title = GLib.path_get_basename(self.archive.path)
            self.headerbar.set_title(title)
            self.props.title = f'{title} - Comipoli' # Application Menu
            self.canvas.is_pdf = self.archive.status == 3
            self.num = 0
            self.change_pixbuf(0)
            # thumbnail
            self.headerbar.grid_button.set_sensitive(False)
            self.thumbnail_array = []
            self.generator = self.create_thumbnail()
            GLib.idle_add(self.idle_function)

    def idle_function(self):
        try:
            next(self.generator)
            return True
        except StopIteration:
            self.headerbar.grid_button.set_sensitive(True)
            return False
        except Exception as e:
            return False

    def create_thumbnail(self):
        for buf in self.archive:
            if self.canvas.is_pdf:
                w, h = buf.get_size()
            else:
                w = buf.get_width()
                h = buf.get_height()
            if w > h:
                aw = self.app.thumbnail_height
                ah = round(self.app.thumbnail_height * h / w)
            else:
                aw = round(self.app.thumbnail_height * w / h)
                ah = self.app.thumbnail_height
            if self.canvas.is_pdf:
                with cairo.ImageSurface(cairo.Format.ARGB32, aw, ah) as surface:
                    cr = cairo.Context(surface)
                    matrix = cairo.Matrix(aw/w, 0, 0, ah/h, 0, 0)
                    cr.transform(matrix)
                    cr.set_source_rgb(1, 1, 1)
                    cr.rectangle(0, 0, w, h)
                    cr.fill()
                    buf.render(cr)
                    pixbuf = Gdk.pixbuf_get_from_surface(surface, 0, 0, aw, ah)
            else:
                pixbuf = buf.scale_simple(aw, ah, GdkPixbuf.InterpType.TILES)
            self.thumbnail_array.append(pixbuf)
            yield 1

    def on_next_info(self, widget, res_id):
        if res_id == Gtk.ResponseType.OK:
            self.set_uri(widget.uri)
        widget.set_visible(False)()

    def on_next_dialog_response(self, dialog, response):
        path = GLib.path_get_dirname(self.archive.path)
        match response:
            case 'next':
                uri = GLib.filename_to_uri(GLib.build_filenamev([path, dialog.cbzs[dialog.next_index]]), None)
                self.set_uri(uri)
            case 'prev':
                uri = GLib.filename_to_uri(GLib.build_filenamev([path, dialog.cbzs[dialog.next_index - 2]]), None)
                self.set_uri(uri)
            case _:
                pass
        dialog.destroy()

    def change_pixbuf(self, nex):
        length = len(self.archive)
        if length == 0:
            return
        num = self.num + nex if self.canvas.spread else self.num + nex - 1
        if length < num + 1 and nex != 50:
            path = GLib.path_get_dirname(self.archive.path)
            etc = GLib.path_get_basename(self.archive.path) # if (zip|rar|7z) ext
            exts = 'cbz'
            if self.app.is_pdf:
                exts += '|pdf'
            if self.app.is_unrar:
                exts += '|cbr'
            if self.app.is_7za:
                exts += '|cb7'
            cbzs = comipoli_numsort.get_file_numsort_list(path, exts, etc)
            l = len(cbzs)
            if l == 0:
                return
            else:
                next_index = cbzs.index(etc) + 1
                if l > next_index:
                    prev_show = False if next_index < 2 else True
                    dlg = ComipolNextDialog(html.escape(cbzs[next_index]), prev_show)
                    dlg.set_transient_for(self)
                    dlg.connect('response', self.on_next_dialog_response)
                    dlg.cbzs = cbzs
                    dlg.next_index = next_index
                    dlg.present()
                elif self.is_fullscreen:
                    self.toggle_fullscreen()
        else:
            if nex == 0:
                # Start Page
                self.canvas.first_page = self.archive[0]
                self.num = 0
                self.canvas.spread = False
                self.canvas.queue_draw()
            elif nex == 50:
                # Last Page
                self.canvas.first_page = self.archive[length - 1]
                self.num = length - 1
                self.canvas.spread = False
                self.canvas.queue_draw()
            elif nex > 0:
                # Next Page
                if length > num:
                    # Set First Page
                    self.canvas.first_page = self.archive[num]
                    if self.is_oblong(self.canvas.first_page):
                        self.canvas.spread = False
                    else:
                        if length > num + 1:
                            # Spread View
                            self.canvas.second_page = self.archive[num + 1]
                            if self.is_oblong(self.canvas.second_page):
                                self.canvas.spread = False
                            else:
                                self.canvas.spread = True
                        else:
                            self.canvas.spread = False
                    self.num = num
                    self.canvas.queue_draw()
            else:
                # Previos Page
                num2 = self.num + nex
                if num2 < 0:
                    self.num = 0
                    self.canvas.first_page = self.archive[0]
                    self.canvas.spread = False
                    self.canvas.queue_draw()
                else:
                    # Check (width > height)
                    pixbuf = self.archive[num2]
                    pixbuf2 = self.archive[num2 + 1] if length > 1 else None
                    if self.is_oblong(pixbuf) or self.is_oblong(pixbuf2):
                        # Single View
                        self.canvas.first_page = pixbuf2
                        self.canvas.spread = False
                        self.num = num2 + 1
                    else:
                        # Spread View
                        self.canvas.first_page = pixbuf
                        self.canvas.second_page = pixbuf2
                        self.canvas.spread = True
                        self.num = num2
                    self.canvas.queue_draw()
            # Draw Page Number
            page_n = f'{self.num + 1}/{length}'
            self.headerbar.set_subtitle(page_n)

    def is_oblong(self, buf):
        if buf == None:
            return True
        if self.archive.status == 3:
            w, h = buf.get_size()
            return w - h > 0
        else:
            return buf.get_width() - buf.get_height() > 0

    def on_lr_button_toggled(self, button):
        if button.props.active:
            self.headerbar.lr_button.set_label('L->R')
            self.upperbar.lr_button.set_label('L->R')
            self.canvas.l_to_r = True
        else:
            self.headerbar.lr_button.set_label('L<-R')
            self.upperbar.lr_button.set_label('L<-R')
            self.canvas.l_to_r = False
        self.canvas.queue_draw()

    def on_mouse_motion(self, widget, x,y):
        if len(self.archive) == 0:
            return
        if self.is_fullscreen:
            if self.menuf.popover.props.visible:
                return
            hb_height = self.headerbar.get_allocated_height()
            if y <= hb_height and self.upperbar.props.visible == False:
                self.upperbar.set_visible(True)()
            elif y > hb_height and self.upperbar.props.visible:
                self.upperbar.set_visible(False)()

    def on_grid_button_clicked(self, button):
        dlg = ComipoliThumbnail(self)
        dlg.set_transient_for(self)
        dlg.present()

    def open_dialog(self):
        ft = Gtk.FileFilter()
        ft.set_name('Comic Book Archive')
        ft.add_mime_type('application/x-cbz')
        ft.add_mime_type('application/vnd.comicbook+zip')
        if self.app.is_pdf:
            ft.add_mime_type('application/pdf')
        if self.app.is_unrar:
            ft.add_mime_type('application/x-cbr')
        if self.app.is_7za:
            ft.add_mime_type('application/x-cb7')
        # 4.10
        dlg = Gtk.FileDialog(default_filter=ft)
        dlg.open(self, None, self.on_open_dlg_callback)

    def on_open_dlg_callback(self, dlg, res):
        try:
            f = dlg.open_finish(res)
            self.set_uri(f.get_uri())
        except Exception as e:
            pass

    def show_fullscreen_menu(self):
        self.menuf.popup()
        return False

    def toggle_fullscreen(self):
        if self.is_fullscreen:
            self.unfullscreen()
            self.is_fullscreen = False
            self.cursor_id = 0
            if self.upperbar.props.visible:
                self.upperbar.set_visible(False)()
        else:
            self.fullscreen()
            self.is_fullscreen = True
            self.cursor_id = 3

    def on_restore_button_clicked(self, widget):
        self.toggle_fullscreen()

    def on_cllick_released(self, widget, n_press, x, y):
        if n_press == 2:
            self.toggle_fullscreen()

    def on_drop_accept(self, target, drop):
        return True

    def on_file_drop(self, target, value, x, y):
        self.set_uri(value.get_files()[0].get_uri())

    def on_window_width_change(self, widget, pspec):
        self.props.application.win_width = self.props.default_width

    def on_window_height_change(self, widget, pspec):
        self.props.application.win_height = self.props.default_height

    def on_window_active(self, widget, pspec):
        self.headerbar.set_active(self.props.is_active)

